library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


entity axis_fifo_cdc is
    
    generic (
        dwidth         : integer := 128;
        uwidth         : integer := 1;
        log2depth      : integer := 9;
        tcq            : integer := 1 );
    port (
        s_clk          : in  std_logic;
        s_reset        : in  std_logic;

        m_clk          : in  std_logic;
        m_reset        : in  std_logic;
        
        s_axis_data    : in  std_logic_vector(dwidth-1 downto 0);
        s_axis_valid   : in  std_logic;
        s_axis_user    : in  std_logic_vector(uwidth-1 downto 0);        
        s_axis_last    : in  std_logic;        
        s_axis_ready   : out std_logic;
        s_axis_empty   : out std_logic;

        m_axis_empty   : out std_logic;
        m_axis_data    : out std_logic_vector(dwidth-1 downto 0);
        m_axis_valid   : out std_logic;
        m_axis_user    : out std_logic_vector(uwidth-1 downto 0);        
        m_axis_last    : out std_logic;        
        m_axis_ready   : in  std_logic );

    attribute keep_hierarchy : STRING;
    attribute keep_hierarchy of axis_fifo_cdc : entity is "yes";    
    
end axis_fifo_cdc;

architecture rtl of axis_fifo_cdc is

    function bin2gray( binslv : std_logic_vector ) return std_logic_vector is
        variable grayslv : std_logic_vector(binslv'range);
    begin
        for i in binslv'range loop
            if i = binslv'left then
                grayslv(i) := binslv(i);
            else
                grayslv(i) := binslv(i) xor binslv(i+1);
            end if;
        end loop;
        return grayslv;
    end function bin2gray;

    function gray2bin( grayslv : std_logic_vector ) return std_logic_vector is
        variable binslv : std_logic_vector(grayslv'range);
    begin
        for i in grayslv'range loop
            if i = grayslv'left then
                binslv(i) := grayslv(i);
            else
                binslv(i) := grayslv(i) xor binslv(i+1);
            end if;
        end loop;
        return binslv;
    end;
    
    
      function f_or (x : std_logic_vector) return std_logic is
    variable result : std_logic := '0';
  begin

    for i in x'range loop
      result := result or x(i);
    end loop;
    return result;
  end f_or;
  
    signal maxspace               : unsigned(log2depth-1 downto 0) := (others => '1');
    
    signal fifo_wren              : std_logic;
    signal fifo_wraddr            : std_logic_vector(log2depth downto 0);
    signal fifo_wraddr_r          : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddr            : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddr_r          : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddrp1          : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddrp1_r        : std_logic_vector(log2depth downto 0);
    signal fifo_wrdata            : std_logic_vector(dwidth+uwidth+1-1 downto 0);
    signal fifo_rddata            : std_logic_vector(dwidth+uwidth+1-1 downto 0);
    signal fifo_rddata_r          : std_logic_vector(dwidth+uwidth+1-1 downto 0);        

    signal fifo_wraddr_gray       : std_logic_vector(log2depth downto 0);
    signal fifo_wraddr_gray_r     : std_logic_vector(log2depth downto 0);
    signal fifo_wraddr_gray_cdc_r : std_logic_vector(log2depth downto 0);
    signal fifo_wraddr_gray_cdc_h : std_logic_vector(log2depth downto 0);
    signal fifo_wraddr_cdc        : std_logic_vector(log2depth downto 0);
    signal fifo_wraddr_cdc_r      : std_logic_vector(log2depth downto 0);

    signal fifo_rdaddr_gray       : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddr_gray_r     : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddr_gray_cdc_r : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddr_gray_cdc_h : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddr_cdc        : std_logic_vector(log2depth downto 0);
    signal fifo_rdaddr_cdc_r      : std_logic_vector(log2depth downto 0);
    attribute async_reg           : string;
    attribute async_reg of fifo_wraddr_gray_cdc_r : signal is "true";
    attribute async_reg of fifo_wraddr_gray_cdc_h : signal is "true";        
    attribute async_reg of fifo_rdaddr_gray_cdc_r : signal is "true";
    attribute async_reg of fifo_rdaddr_gray_cdc_h : signal is "true";       
            
    signal fifo_rdy               : std_logic;
    signal fifo_read              : std_logic;
    signal fifo_read_r            : std_logic;
    signal fifo_read_h            : std_logic;
    signal fifo_ready             : std_logic;
    signal fifo_ready_r           : std_logic;    

    signal fifo_space             : unsigned(log2depth-1 downto 0);
    signal fifo_space_r, fifo_space_m           : unsigned(log2depth-1 downto 0);
    signal fifo_dvalid            : std_logic;
    signal fifo_dvalid_r          : std_logic;
    signal fifo_restrobe          : std_logic;
    signal fifo_restrobe_r        : std_logic;

attribute syn_radhardlevel :  string;
attribute syn_radhardlevel of rtl : architecture is "tmr";

begin

    fifo_wren        <= s_axis_valid and fifo_ready_r;
    fifo_wraddr      <= std_logic_vector((unsigned(fifo_wraddr_r) + 1)) when fifo_wren = '1' else fifo_wraddr_r;
    fifo_wrdata      <= (s_axis_last & s_axis_user & s_axis_data);
    fifo_wraddr_gray <= bin2gray(fifo_wraddr);
    fifo_rdaddr_cdc  <= gray2bin(fifo_rdaddr_gray_cdc_h);
    fifo_space       <= maxspace + unsigned(fifo_rdaddr_cdc_r(log2depth-1 downto 0)) - unsigned(fifo_wraddr(log2depth-1 downto 0));    
    fifo_ready       <= '1' when f_or(std_logic_vector(fifo_space)) = '1' else '0';
    fifo_space_m      <= maxspace + unsigned(fifo_rdaddr(log2depth-1 downto 0)) - unsigned(fifo_wraddr_cdc_r(log2depth-1 downto 0));
    mem_i : entity work.sdpram_cdc_r
        generic map (
            width      => dwidth + uwidth + 1,
            log2depth  => log2depth,           
            tcq        => tcq )
        port map (
            wrclk      => s_clk,
            rdclk      => m_clk,            
            wren       => fifo_wren,
            wraddr     => fifo_wraddr_r(log2depth-1 downto 0),
            rdaddr     => fifo_rdaddr(log2depth-1 downto 0),
            wrdata     => fifo_wrdata,
            rddata     => fifo_rddata );

    fifo_wraddr_cdc  <= gray2bin(fifo_wraddr_gray_cdc_h);
    fifo_rdy         <= '1' when fifo_wraddr_cdc_r /= fifo_rdaddrp1_r else '0';
    fifo_read        <= fifo_rdy and ((not  (fifo_read_r  or fifo_dvalid  or fifo_restrobe_r))                   or
                                      (      fifo_read_r xor fifo_dvalid xor fifo_restrobe_r)                    or
                                      ((not (fifo_read_r and fifo_dvalid and fifo_restrobe_r)) and m_axis_ready));
    fifo_rdaddr      <= std_logic_vector(unsigned(fifo_rdaddr_r)   + 1) when fifo_read = '1' else fifo_rdaddr_r;
    fifo_rdaddrp1    <= std_logic_vector(unsigned(fifo_rdaddrp1_r) + 1) when fifo_read = '1' else fifo_rdaddrp1_r;
    fifo_rdaddr_gray <= bin2gray(fifo_rdaddr);

    fifo_dvalid      <= '1' when (fifo_read_h or fifo_dvalid_r) = '1' else '0';
    fifo_restrobe    <= '1' when ((fifo_dvalid and (not m_axis_ready)) or (fifo_restrobe_r and (not m_axis_ready))) = '1' else '0';

    
    process (s_clk) is
    begin
        if rising_edge(s_clk) then
            if s_reset = '1' then
                fifo_ready_r           <= '0'                    after (tcq) * 1 ps;                
                fifo_wraddr_r          <= (others => '0')        after (tcq) * 1 ps;
                fifo_wraddr_gray_r     <= (others => '0')        after (tcq) * 1 ps;
                fifo_rdaddr_gray_cdc_r(log2depth-2 downto 0) <= (others => '0') after (tcq) * 1 ps;
                fifo_rdaddr_gray_cdc_r(log2depth-1)          <= ('1')           after (tcq) * 1 ps;
                fifo_rdaddr_gray_cdc_h(log2depth-2 downto 0) <= (others => '0') after (tcq) * 1 ps;
                fifo_rdaddr_gray_cdc_h(log2depth-1)          <= ('1')           after (tcq) * 1 ps;                
                fifo_rdaddr_cdc_r      <= (others => '1')        after (tcq) * 1 ps;
            else
                fifo_ready_r           <= fifo_ready             after (tcq) * 1 ps; 
                fifo_wraddr_r          <= fifo_wraddr            after (tcq) * 1 ps;
                fifo_wraddr_gray_r     <= fifo_wraddr_gray       after (tcq) * 1 ps;
                fifo_rdaddr_gray_cdc_r <= fifo_rdaddr_gray_r     after (tcq) * 1 ps;
                fifo_rdaddr_gray_cdc_h <= fifo_rdaddr_gray_cdc_r after (tcq) * 1 ps;                
                fifo_rdaddr_cdc_r      <= fifo_rdaddr_cdc        after (tcq) * 1 ps;
            end if;
        end if;
    end process;
    
    
    process (m_clk) is
    begin
        if rising_edge(m_clk) then
            if m_reset = '1' then
                fifo_wraddr_gray_cdc_r <= (others => '0')        after (tcq) * 1 ps;
                fifo_wraddr_gray_cdc_h <= (others => '0')        after (tcq) * 1 ps;
                fifo_wraddr_cdc_r      <= (others => '0')        after (tcq) * 1 ps;
                fifo_rdaddr_r          <= (others => '1')        after (tcq) * 1 ps;
                fifo_rdaddr_gray_r(log2depth-2 downto 0)     <=  (others => '0') after (tcq) * 1 ps;
                fifo_rdaddr_gray_r(log2depth-1)              <=  ('1')           after (tcq) * 1 ps;                   
                fifo_rdaddrp1_r        <= (others => '0')        after (tcq) * 1 ps;
                fifo_dvalid_r          <= '0'                    after (tcq) * 1 ps;
                fifo_restrobe_r        <= '0'                    after (tcq) * 1 ps;
                fifo_read_r            <= '0'                    after (tcq) * 1 ps;
                fifo_read_h            <= '0'                    after (tcq) * 1 ps;
                m_axis_empty <= '1';
            else
                fifo_wraddr_gray_cdc_r <= fifo_wraddr_gray_r     after (tcq) * 1 ps;
                fifo_wraddr_gray_cdc_h <= fifo_wraddr_gray_cdc_r after (tcq) * 1 ps;                
                fifo_wraddr_cdc_r      <= fifo_wraddr_cdc        after (tcq) * 1 ps;
                fifo_rdaddr_r          <= fifo_rdaddr            after (tcq) * 1 ps;
                fifo_rdaddr_gray_r     <= fifo_rdaddr_gray       after (tcq) * 1 ps;
                fifo_rdaddrp1_r        <= fifo_rdaddrp1          after (tcq) * 1 ps;
                fifo_restrobe_r        <= fifo_restrobe          after (tcq) * 1 ps;                
                fifo_read_r            <= fifo_read              after (tcq) * 1 ps;
                fifo_read_h            <= fifo_read_r            after (tcq) * 1 ps;

                if (fifo_restrobe_r and ((not m_axis_ready) or fifo_dvalid)) = '1' then
                    fifo_dvalid_r      <= fifo_dvalid            after (tcq) * 1 ps;
                else
                    fifo_dvalid_r      <= '0'                    after (tcq) * 1 ps;
                end if;

                if (fifo_restrobe and (not fifo_restrobe_r)) = '1' then
                    fifo_rddata_r      <= fifo_rddata            after (tcq) * 1 ps;
                end if;
                if (m_axis_ready = '1' and fifo_read = '0') then
                  
                     m_axis_empty <= '1';
                elsif ( (m_axis_ready = '1' and fifo_read = '1') or fifo_read = '1') then
                    m_axis_empty <= '0';
            end if;
                
              
            end if;            
        end if;
    end process;

    -- outputs
    s_axis_ready  <= fifo_ready_r;
    s_axis_empty <= '1' when fifo_space = maxspace-1 else '0';
    m_axis_valid  <= fifo_dvalid or fifo_restrobe_r;

     m_axis_data   <= fifo_rddata_r(dwidth-1        downto 0)      when fifo_restrobe_r = '1' else fifo_rddata(dwidth-1        downto 0);
    m_axis_user   <= fifo_rddata_r(dwidth+uwidth-1 downto dwidth) when fifo_restrobe_r = '1' else fifo_rddata(dwidth+uwidth-1 downto dwidth);
    m_axis_last   <= fifo_rddata_r(dwidth+uwidth)                 when fifo_restrobe_r = '1' else fifo_rddata(dwidth+uwidth);   
    

end architecture;


