
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--library synplify;
--use synplify.attributes.all;
--use synplify.all;

library work;
use work.wishbone_pkg.all;
--
--library smartfusion2;
--use smartfusion2.all;


entity ueth_dma is

  
  port (
    clk_i : in std_logic;
    rst_i : in std_logic;

    rx_en_o : out std_logic;
    tx_en_o : out std_logic;
    rx_ready_o : out std_logic;
    
    rx_done_i      : in std_logic;
    tx_done_i      : in std_logic;
    rx_overflow_i  : in std_logic;
    tx_underflow_i : in std_logic;

    tx_pending_o : out std_logic;
    rx_pending_o : out std_logic;

    tx_fifo_data_o  : out std_logic_vector (35 downto 0);
    tx_fifo_valid_o : out std_logic;
    tx_fifo_ready_i : in  std_logic;
   -- tx_fifo_last_o : out std_logic;
 
    rx_fifo_data_i  : in  std_logic_vector (35 downto 0);
    rx_fifo_valid_i : in  std_logic;
    rx_fifo_ready_o : out std_logic;
    rx_fifo_eof_i : in std_logic;

    irq_o : out std_logic;

    mii_rd_o   : out std_logic;
    mii_done_i : in  std_logic;
    mii_addr_o : out std_logic_vector (4 downto 0);
    mii_reg_i  : in  std_logic_vector (15 downto 0);

  
    rst_mac_o : out std_logic;

    slave_i : in  t_wishbone_slave_in;
    slave_o : out t_wishbone_slave_out;

    master_o : out t_wishbone_master_out;
    master_i : in  t_wishbone_master_in

    );
end entity ueth_dma;


architecture rtl of ueth_dma is

  type r_descriptor is record
    int_flag        : std_logic;
    size            : unsigned (10 downto 0);
    addr            : std_logic_vector (15 downto 0);
    busy            : std_logic;
    fifo_flow_error : std_logic;
  end record r_descriptor;

  function f_byte_count (byte_sel : std_logic_vector (3 downto 0)) return integer is
  begin
    case (byte_sel) is
      when "0001" => return 1;
      when "0011" => return 2;
      when "0111" => return 3;
      when "1111" => return 4;
      when others => return 0;
    end case;
  end f_byte_count;

  function f_byte_sel (byte_counter : unsigned (1 downto 0)) return std_logic_vector is
  begin
    case (to_integer (byte_counter)) is
      when 0      => return "1111";
      when 1      => return "0001";
      when 2      => return "0011";
      when 3      => return "0111";
      when others => return "XXXX";
    end case;
  end f_byte_sel;


  constant c_ETHERTYPE_POWERLINK : std_logic_vector (15 downto 0) := x"ab88";
  constant rx_descriptor0_addr : std_logic_vector (15 downto 0) := x"f400";
  constant rx_descriptor1_addr : std_logic_vector (15 downto 0) := x"f9ec";

  type t_fsm_rx is (idle, setup, read_word, write_word);
  type t_fsm_tx is (idle, read_word, write_word, write_partial_word, wait_tx_done);
  type t_fsm_dma is (idle, transfer_read, transfer_write);

  signal fsm_rx  : t_fsm_rx  := idle;
  signal fsm_tx  : t_fsm_tx  := idle;
  signal fsm_dma : t_fsm_dma := idle;

  signal last_word : std_logic;
  signal rx_descriptor0, rx_descriptor1, tx_descriptor              : r_descriptor;
  signal tx_enable, rx_enable, mii_scrubber_enable                  : std_logic;
  signal peding_tx, pending_rx                                      : std_logic;
  signal dma_write, dma_read                                        : std_logic;
  signal read_addr, write_addr                                      : unsigned (15 downto 0);
  signal dma_write_bsel                                             : std_logic_vector (3 downto 0);
  signal tx_pkt_sent_p, rx_pkt_recv_p                               : std_logic;
  signal tx_bytes, rx_bytes                                         : unsigned (10 downto 0);
  signal dma_read_req, dma_read_done, dma_write_req, dma_write_done : std_logic;
  signal data_tx, data_rx                                           : std_logic_vector (31 downto 0);
  signal rx_desc_taken                                              : std_logic;
  signal rx_ethertype                                               : std_logic_vector(15 downto 0);

signal mii_read_p, mii_done : std_logic;
signal mii_reg : std_logic_vector (15 downto 0);

signal dma_read_write : std_logic;

signal mac_rx_timeout : unsigned (5 downto 0);

attribute syn_encoding : string; 
attribute syn_encoding of fsm_rx, fsm_tx, fsm_dma : signal is "safe,onehot";

attribute syn_radhardlevel : string;
attribute syn_radhardlevel of rtl : architecture is "tmr";
begin

  irq_o <= rx_descriptor0.int_flag or rx_descriptor1.int_flag or tx_descriptor.int_flag;

  rx_en_o <= rx_enable;
  tx_en_o <= tx_enable;

  rx_ready_o <= (rx_descriptor0.busy or rx_descriptor1.busy) and not rx_pending_o;

  rx_descriptor0.addr <= rx_descriptor0_addr;
  rx_descriptor1.addr <= rx_descriptor1_addr;

  tx_descriptor.fifo_flow_error <= tx_underflow_i;

  slave_o.stall <= (not slave_o.ack) and slave_i.cyc and slave_i.stb;

  process (clk_i)
  begin

    if rising_edge (clk_i) then
      if (rst_i) then
        tx_enable               <= '0';
        rx_enable               <= '0';
        rx_descriptor0.busy     <= '0';
        rx_descriptor1.busy     <= '0';
        tx_descriptor.busy      <= '0';
        rx_descriptor0.int_flag <= '0';
        rx_descriptor1.int_flag <= '0';
        tx_descriptor.int_flag  <= '0';
        mii_rd_o <= '0';
        rst_mac_o <= '1';


      else
        slave_o.ack <= '0';
        mii_rd_o <= '0';
        if (slave_i.stb = '1' and slave_i.cyc = '1' and slave_o.ack = '0') then
          slave_o.ack <= '1';
          if (slave_i.adr(4 downto 2) = "000") then
            slave_o.dat (31)           <= rx_descriptor0.busy;
            slave_o.dat (30)           <= rx_descriptor0.fifo_flow_error;
            slave_o.dat (29)           <= rx_descriptor0.int_flag;
            slave_o.dat (26 downto 16) <= std_logic_vector(rx_descriptor0.size);
            slave_o.dat (15 downto 0)  <= (rx_descriptor0.addr);
            rx_descriptor0.int_flag    <= '0';
            if (slave_i.we = '1') then
              --  rx_descriptor0.int_flag <=   slave_o.dat (29);
              if (rx_descriptor0.busy = '0') then
                rx_descriptor0.busy <= slave_i.dat(31);
              --rx_descriptor.size <= unsigned(slave_i.dat (10 downto 0));
              end if;
            end if;
          elsif (slave_i.adr(4 downto 2) = "001") then
            slave_o.dat (31)           <= rx_descriptor1.busy;
            slave_o.dat (30)           <= rx_descriptor1.fifo_flow_error;
            slave_o.dat (29)           <= rx_descriptor1.int_flag;
            slave_o.dat (26 downto 16) <= std_logic_vector(rx_descriptor1.size);
            slave_o.dat (15 downto 0)  <= (rx_descriptor1.addr);
            rx_descriptor1.int_flag    <= '0';
            if (slave_i.we = '1') then
              -- rx_descriptor1.int_flag <=   slave_o.dat (29);
              if (rx_descriptor1.busy = '0') then
                rx_descriptor1.busy <= slave_i.dat(31);
              --rx_descriptor.size <= unsigned(slave_i.dat (10 downto 0));
              end if;
            end if;
          elsif (slave_i.adr(4 downto 2) = "010") then
            slave_o.dat (31)           <= tx_descriptor.busy;
            slave_o.dat (30)           <= tx_descriptor.fifo_flow_error;
            slave_o.dat (29)           <= tx_descriptor.int_flag;
            slave_o.dat (26 downto 16) <= std_logic_vector(tx_descriptor.size);  --(
            slave_o.dat (15 downto 0)  <= tx_descriptor.addr;  -- 
            tx_descriptor.int_flag     <= '0';
            if (slave_i.we = '1') then
              if (tx_descriptor.busy = '0') then
                tx_descriptor.busy <= slave_i.dat(31);
                tx_descriptor.size <= unsigned(slave_i.dat (26 downto 16));
                tx_descriptor.addr <= slave_i.dat(15 downto 0);
              end if;
            end if;
          elsif (slave_i.adr(4 downto 2) = "011") then
         
            slave_o.dat (31 downto 3) <= (others => '0');
            slave_o.dat (2 downto 0)  <=  rst_mac_o & tx_enable & rx_enable;
            
            if (slave_i.we = '1') then
              rst_mac_o <= slave_i.dat(3); 
              tx_enable           <= slave_i.dat(1);
              rx_enable           <= slave_i.dat(0);
              if (slave_i.dat(2) = '1') then
                tx_enable <= '0';
                rx_enable <= '0';
              end if;
            end if;
          elsif (slave_i.adr(4 downto 2) = "100") then
              if (slave_i.we = '1') then
                mii_addr_o <= slave_i.dat(4 downto 0);
                mii_rd_o <= '1';
            end if;
          elsif (slave_i.adr(4 downto 2) = "101") then
            slave_o.dat (31)           <= mii_done_i;
            slave_o.dat (30 downto 16) <= (others => '0');
            slave_o.dat (15 downto 0)  <= mii_reg_i;
          end if;
        end if;
          if (rx_pkt_recv_p = '1' and rx_desc_taken = '0') then
            rx_descriptor0.busy     <= '0';
            rx_descriptor0.int_flag <= '1';
          end if;
          if (rx_pkt_recv_p = '1' and rx_desc_taken = '1') then
            rx_descriptor1.busy     <= '0';
            rx_descriptor1.int_flag <= '1';
          end if;
          if (tx_pkt_sent_p = '1') then
            tx_descriptor.busy     <= '0';
            tx_descriptor.int_flag <= '1';
          end if;
 
 
        end if;
      end if;
    end process;

-- rx flux

      process (clk_i)
      begin
        if rising_edge (clk_i) then
          if (rst_i) then
            fsm_rx                         <= idle;
            rx_descriptor0.fifo_flow_error <= '0';
            rx_descriptor1.fifo_flow_error <= '0';
            rx_descriptor0.size            <= (others => '0');
            rx_descriptor1.size            <= (others => '0');
            dma_write_req                  <= '0';
            rx_pending_o <= '0';
            mac_rx_timeout <= (others => '0');
          else
            rx_pkt_recv_p <= '0';
            case (fsm_rx) is
              when idle =>
                rx_pending_o <= '0';
                rx_fifo_ready_o <= '0';
                if (rx_fifo_valid_i = '1' and rx_enable = '1' and rx_done_i = '0' and ((rx_descriptor0.busy = '1' or rx_descriptor1.busy = '1') and rx_pkt_recv_p = '0')) then
                  fsm_rx <= read_word;
                  if (rx_descriptor0.busy = '1') then
                    write_addr    <= unsigned(rx_descriptor0.addr);
                    rx_desc_taken <= '0';
                  else
                    write_addr    <= unsigned(rx_descriptor1.addr);
                    rx_desc_taken <= '1';
                  end if;
                  rx_pending_o <= '1';
                  rx_bytes     <= (others => '0');
                end if;

              when read_word =>
                rx_fifo_ready_o <= '0';
                if (rx_fifo_valid_i = '1' and rx_fifo_eof_i = '0') then
                  rx_fifo_ready_o <= '1';
                  
                  data_rx         <= rx_fifo_data_i (31 downto 0);
                  dma_write_req   <= '1';
                  fsm_rx          <= write_word;
                  dma_write_bsel  <= rx_fifo_data_i (35 downto 32);
                  rx_bytes        <= rx_bytes + f_byte_count (rx_fifo_data_i (35 downto 32));
                 
                  if (rx_bytes = to_unsigned(12, 11)) then
                    rx_ethertype <= rx_fifo_data_i (15 downto 0);
                  end if;
                end if;
                 if (rx_fifo_valid_i = '1' and rx_fifo_eof_i = '1') then
                   fsm_rx        <= idle;
                    rx_fifo_ready_o <= '1';
                 -- if (rx_ethertype = c_ETHERTYPE_POWERLINK) then --If a tree falls in a forest and no one is around to hear it, does it make a sound?
                  rx_pkt_recv_p <= '1';
                --  end if;
                  if (rx_desc_taken = '0') then
                    rx_descriptor0.size            <= rx_bytes;
                    rx_descriptor0.fifo_flow_error <= rx_overflow_i;
                  else
                    rx_descriptor1.size            <= rx_bytes;
                    rx_descriptor1.fifo_flow_error <= rx_overflow_i;
                  end if;
               end if;

                -- timeout if EOF flag is corrupted    
               if  (rx_fifo_valid_i = '1') then
                 mac_rx_timeout <= (others => '0');
               else
                 mac_rx_timeout <= mac_rx_timeout + 1;
                 if (mac_rx_timeout (5)) then
                   fsm_rx <= idle;
                 end if;
               end if;
                 

              when write_word =>
                rx_fifo_ready_o <= '0';
                if (dma_write_done = '1') then
                  dma_write_req <= '0';
                  write_addr    <= write_addr + 4;
                  fsm_rx        <= read_word;
               
                 
           
                end if;
              when others =>
                fsm_rx <= idle;

            end case;
          end if;
        end if;
      end process;

      -- tx flux

      process (clk_i)
      begin
        if rising_edge (clk_i) then
          if (rst_i) then
            fsm_tx       <= idle;
            dma_read_req <= '0';
            tx_fifo_valid_o <= '0';
          else
            tx_pkt_sent_p <= '0';
            case (fsm_tx) is
              when idle =>
                tx_pending_o    <= '0';
                if (tx_enable = '1' and (tx_descriptor.busy = '1' and tx_pkt_sent_p = '0')) then
                  fsm_tx    <= read_word;
                  read_addr <= unsigned(tx_descriptor.addr);
                  tx_bytes  <= (others => '0');

                end if;

              when read_word =>

              
                if (tx_bytes (10 downto 2) >= tx_descriptor.size (10 downto 2)) then
                  if (tx_descriptor.size (1 downto 0) = 0) then
                    fsm_tx        <= idle;
                    tx_pkt_sent_p <= '1';
                    tx_pending_o  <= '0';

                  else
                    dma_read_req <= '1';
                    fsm_tx       <= write_partial_word;
                  end if;
                elsif (tx_fifo_ready_i = '1') then
                  dma_read_req <= '1';
                  fsm_tx       <= write_word;

                end if;

              when write_word =>
                --dma_read_req <= '0';
                --tx_fifo_last_o  <= '0';
                if (dma_read_done = '1' ) then
                  dma_read_req    <= '0';
                
                tx_fifo_valid_o <= '1';
                tx_fifo_data_o  <= "1111" & data_tx;
                read_addr       <= read_addr + 4;
                tx_bytes        <= tx_bytes + 4;
                end if;
                if (tx_fifo_valid_o = '1' and tx_fifo_ready_i = '1') then
                   fsm_tx          <= read_word;
                   tx_fifo_valid_o <= '0';
                end if;
              
                
                if (tx_bytes (4) = '1') then
                  tx_pending_o <= '1';
                  
              end if;
              when write_partial_word =>
                --dma_read_req <= '0';
                if (dma_read_done = '1') then
                  dma_read_req    <= '0';
                
                 
                  tx_fifo_valid_o <= '1';
                  tx_fifo_data_o  <= f_byte_sel(tx_descriptor.size(1 downto 0)) & data_tx;
                  read_addr       <= read_addr + 4;
                  tx_bytes        <= tx_bytes + tx_descriptor.size (1 downto 0);
                 end if;
                 if (tx_fifo_valid_o = '1' and tx_fifo_ready_i = '1') then 
                  fsm_tx          <= idle;
                  tx_pkt_sent_p   <= '1';
                  tx_pending_o    <= '0';
                  tx_fifo_valid_o <= '0';
                end if;
              

              when wait_tx_done =>
                if (tx_done_i = '1') then
                  fsm_tx <= idle;
                end if;
                
            when others =>
              fsm_tx <= idle;

            end case;
          end if;
        end if;
      end process;

      master_o.adr(31 downto 16) <= x"8500";
dma_write <=  dma_write_req and not dma_write_done;
dma_read <= dma_read_req and not dma_read_done;
      process (clk_i)
      begin
        if rising_edge(clk_i) then
          if (rst_i) then
            fsm_dma        <= idle;
            master_o.cyc   <= '0';
            master_o.stb   <= '0';
            dma_read_done  <= '0';
            dma_write_done <= '0';
            dma_read_write <= '0'; -- read = 1
          else
            case (fsm_dma) is

              when idle =>
                dma_write_done <= '0';
                dma_read_done  <= '0';
                if (dma_write = '1' and dma_read = '1') then
                  dma_read_write <= not dma_read_write;
                  if (dma_read_write = '1') then
                  fsm_dma   <= transfer_write;
                  else
                  fsm_dma   <= transfer_read;
                end if;
              elsif (dma_write = '1' or dma_read = '1') then
                  if (dma_write = '1') then
                  fsm_dma   <= transfer_write;
                  dma_read_write <= '0';
                else
                  fsm_dma <= transfer_read;
                  dma_read_write <= '1';
              end if;
                  

                end if;
              when transfer_read =>
                master_o.cyc               <= '1';
                master_o.stb               <= '1';
                master_o.we                <= '0';
                master_o.sel               <= "1111"        ;
                master_o.adr (15 downto 0) <= std_logic_vector(read_addr); 
                master_o.dat               <= data_rx;
          
                
                data_tx <= master_i.dat;
            
                if (master_i.ack = '1') then
                  master_o.cyc   <= '0';
                  master_o.stb   <= '0';
    
                  dma_read_done  <= '1';
                  fsm_dma        <= idle;

                end if;

              when transfer_write =>
                master_o.cyc               <= '1';
                master_o.stb               <= '1';
                master_o.we                <= '1';
                master_o.sel               <=  dma_write_bsel;
                master_o.adr (15 downto 0) <= std_logic_vector(write_addr);
                master_o.dat               <= data_rx;
          
          
                if (master_i.ack = '1') then
                  master_o.cyc   <= '0';
                  master_o.stb   <= '0';
                  dma_write_done <= '1';
                  
                  fsm_dma        <= idle;

                end if;

              when others =>
                fsm_dma <= idle;
            end case;
          end if;
        end if;
      end process;



    end architecture;
