
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--library synplify;
--use synplify.attributes.all;
--use synplify.all;

library work;
use work.wishbone_pkg.all;
--
--library smartfusion2;
--use smartfusion2.all;


entity ueth_mac is

  
  port (
  
  
    clk_i : in std_logic;
    rst_i : in std_logic;

    rst_phy_o : out std_logic;
    phy_rxen_o: out std_logic;
    phy_txen_o: out std_logic;

    mac_rxd_i : in std_logic_vector (3 downto 0);
    mac_txd_o : out std_logic_vector (3 downto 0);
    mac_rx_valid_i : in std_logic;
    mac_tx_valid_o : out std_logic;
    mac_tx_ready_i : in std_logic;
    mac_tx_eof_o : out std_logic;
    mac_rx_eof_i : in std_logic;

    irq_o : out std_logic;

    slave_i : in  t_wishbone_slave_in;
    slave_o : out t_wishbone_slave_out;

    master_o : out t_wishbone_master_out;
    master_i : in  t_wishbone_master_in;    
    
    -- MIIM
    phy_mdc_o : out std_logic;
    phy_md_i : in std_logic;
    phy_md_o : out std_logic;
    phy_oe_o : out std_logic
 
    );
end entity ueth_mac;


architecture rtl of ueth_mac is

type rx_fsm_states is (idle, preamble_check, wait_sfd, data, eof); 
type tx_fsm_states is (idle, preamble, sfd, data, fetch_data, crc_wait, crc, gap);

signal rx_fsm : rx_fsm_states := idle;
signal tx_fsm : tx_fsm_states := idle;

signal rx_overflow, rx_done : std_logic;
signal rx_nibble_counter : unsigned (2 downto 0);

signal tx_underflow, tx_done : std_logic;
signal tx_nibble_counter : unsigned (3 downto 0);
signal rx_ready : std_logic;
signal tx_en, rx_en, tx_pending, rx_pending : std_logic;
signal rst_mac : std_logic;
signal tx_fifo_data_dma, rx_fifo_data_dma : std_logic_vector (35 downto 0);
signal tx_fifo_valid_dma, tx_fifo_ready_dma, rx_fifo_valid_dma, rx_fifo_ready_dma  : std_logic;
signal tx_fifo_last_dma : std_logic;
signal rxd_word, txd_word : std_logic_vector (35 downto 0);
signal rxd_word_valid, txd_word_valid, txd_word_ready, rxd_word_ready : std_logic;
signal rxd_word_eof, txd_word_eof : std_logic;
signal rx_fifo_eof_dma : std_logic;
signal mii_rd, mii_done : std_logic;
signal mii_addr :  std_logic_vector (4 downto 0);
signal mii_reg : std_logic_vector (15 downto 0);

signal crc_tx, crc_rx, crc_tx_reversed : std_logic_vector (31 downto 0);
signal rst_crc_rx, rst_crc_tx : std_logic;
signal en_crc_tx : std_logic;

signal mac_rx_timeout : integer range 0 to 7;


attribute syn_encoding : string; 
attribute syn_encoding of rx_fsm, tx_fsm : signal is "safe,onehot";



attribute syn_radhardlevel : string;
attribute syn_radhardlevel of rtl : architecture is "tmr";

function reverse_vector (a: in std_logic_vector)
return std_logic_vector is
  variable result: std_logic_vector(a'RANGE);
  alias aa: std_logic_vector(a'REVERSE_RANGE) is a;
begin
  for i in aa'RANGE loop
    result(i) := aa(i);
  end loop;
  return result;
end; -- function reverse_any_vector


begin

phy_rxen_o <= rx_en;
phy_txen_o <= tx_en;
rst_phy_o <= rst_mac;

 -- fsm rx
  
  process (clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_i = '1' or rst_mac = '1' then
        rx_fsm <= idle;
        rx_done <= '0';
        rx_overflow <= '0';
        rxd_word_valid <= '0';
        mac_rx_timeout <= 0;
      else
      
      
      case (rx_fsm) is
        when idle =>
           rx_nibble_counter <= (others => '0');
           if (rxd_word_ready = '1') then
           rxd_word_valid <= '0';
         end if;
          if (rx_en = '1'  and rx_ready = '1' and mac_rx_valid_i = '1' and mac_rxd_i = x"5") then
            rx_fsm <= preamble_check;
            rx_overflow <= '0';
            rx_done <= '0';
            rxd_word_eof <= '0';
          end if;
        when preamble_check =>
          if (mac_rx_valid_i = '1' and mac_rxd_i /= x"5") then
            rx_fsm <= idle;
          elsif (mac_rx_valid_i = '1' and mac_rxd_i = x"5") then
            rx_nibble_counter <= rx_nibble_counter + 1;
            if (rx_nibble_counter = "111") then
              rx_fsm <= wait_sfd;
            end if;
          end if;
        when wait_sfd =>
          if (mac_rx_valid_i = '1' and mac_rxd_i = x"D") then
            rx_fsm <= data;
            rx_nibble_counter <= (others => '0');
          elsif  (mac_rx_valid_i = '1' and mac_rxd_i /= x"D" and mac_rxd_i /= x"5") then
            rx_fsm <= idle;
          end if;
        when data =>
          rxd_word_valid <= '0';
          if (mac_rx_valid_i = '1' and mac_rx_eof_i = '0') then
            rx_nibble_counter <= rx_nibble_counter + 1;
            case (to_integer(rx_nibble_counter)) is
              
                when 7 => rxd_word (31 downto 28) <= mac_rxd_i;
                          rxd_word (35) <= '1';
                when 6 => rxd_word (27 downto 24) <= mac_rxd_i;
                when 5 => rxd_word  (23 downto 20) <= mac_rxd_i;
                          rxd_word (34) <= '1';
                when 4 => rxd_word  (19 downto 16) <= mac_rxd_i;
                when 3 => rxd_word  (15 downto 12) <= mac_rxd_i;
                          rxd_word (33) <= '1';
                when 2 => rxd_word (11 downto 8) <= mac_rxd_i;
                when 1 => rxd_word (7 downto 4) <= mac_rxd_i;
                          rxd_word (32) <= '1';
                when 0 => rxd_word (3 downto 0)   <= mac_rxd_i;
                          rxd_word (35 downto 32) <= "0000";
                when others => rxd_word <= (others => 'X');
               end case;                                                            
          

          if (rx_nibble_counter = "111") then
             rxd_word_valid <= '1';
             if (rxd_word_ready = '0') then
               rx_overflow <= '1';
             end if;
           end if;
          end if;
          if (mac_rx_valid_i = '1' and mac_rx_eof_i = '1') then
             rx_fsm <= eof;
             if (to_integer(rx_nibble_counter) /= 0 ) then
              rxd_word_valid <= '1';
            end if;
          end if;

          -- timeout if EOF flag is corrupted
          if (mac_rx_valid_i = '1') then
            mac_rx_timeout <= 0;
          else
            mac_rx_timeout <= mac_rx_timeout + 1;
            if (mac_rx_timeout > 4) then
              rx_fsm <= eof;
            end if;
          end if;
          
      when eof =>
             if (rxd_word_ready = '1') then
              rx_done <= '1';
              rx_fsm <= idle;
              rxd_word_valid <= '1';
              rxd_word_eof <= '1';
             end if;
          
   

      when others =>
        rx_fsm <= idle;                         
   end case;
      end if;
      end if;
      end process;
           
                               
                               

-- fsm tx
--


process (clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_i = '1' or rst_mac = '1' then
        tx_fsm <= idle;
        mac_tx_valid_o <= '0';
        txd_word_ready <= '0';
        mac_tx_eof_o <= '0';
      else
      case (tx_fsm) is
        when idle =>
          tx_underflow <= '0';
          mac_txd_o <= x"0";
            if (tx_en = '1' and tx_pending = '1' and txd_word_valid = '1') then
              tx_fsm <= preamble;
              tx_nibble_counter <= (others => '0');
            
            end if;
        when preamble =>
          mac_txd_o <= x"5";
          mac_tx_valid_o <= '1';
          if (mac_tx_valid_o = '1' and mac_tx_ready_i = '1') then
            tx_nibble_counter <= tx_nibble_counter + 1;
          end if;
          if (mac_tx_ready_i = '1' and tx_nibble_counter = 14) then
              tx_fsm <= sfd;
              mac_txd_o <= x"D";
          end if;
        when sfd =>        
          if (mac_tx_ready_i = '1') then
            tx_fsm <= data;
            mac_tx_valid_o <= '0';
           
          end if;
          tx_nibble_counter <= (others => '0');
          
        when data =>
          
          txd_word_ready <= '0';
          
        
         if (mac_tx_valid_o = '0' or (mac_tx_ready_i = '1' and mac_tx_valid_o = '1')) then
           mac_tx_valid_o <= '1';    
            case (to_integer(tx_nibble_counter (2 downto 0))) is
              when 0 => mac_txd_o <= txd_word (3 downto 0);
              when 1 => mac_txd_o <= txd_word (7 downto 4);
                        if (txd_word(33) = '0') then
                          tx_fsm <= crc_wait;
                          txd_word_ready <= '1';
                        end if;
              when 2 => mac_txd_o <= txd_word (11 downto 8);
              when 3 => mac_txd_o <= txd_word (15 downto 12);
                        if (txd_word(34) = '0') then
                          tx_fsm <= crc_wait;
                          txd_word_ready <= '1';
                        end if;
              when 4 => mac_txd_o <= txd_word (19 downto 16);
              when 5 => mac_txd_o <= txd_word (23 downto 20);
                        if (txd_word(35) = '0') then
                          tx_fsm <= crc_wait;
                          txd_word_ready <= '1';
                        end if;
              when 6 => mac_txd_o <= txd_word (27 downto 24);

              when 7 => mac_txd_o <= txd_word (31 downto 28);
                        txd_word_ready <= '1';
                        tx_fsm <= fetch_data;
             when others => mac_txd_o <= (others => 'X');
                            txd_word_ready <= 'X';
                      
            end case;
         end if;
          if  (mac_tx_valid_o = '0' or (mac_tx_ready_i = '1' and mac_tx_valid_o = '1')) then
            
            tx_nibble_counter <= tx_nibble_counter + 1;
          end if;
          

          
      when fetch_data =>
        tx_nibble_counter <= (others => '0');
        if (mac_tx_valid_o = '1' and mac_tx_ready_i = '1') then
          mac_tx_valid_o <= '0';
        end if;
        
        txd_word_ready <= '0';
        if (txd_word_valid = '0') then
            if (tx_pending = '1') then
              tx_underflow <= '1';
          elsif mac_tx_valid_o = '0' then
              tx_fsm <= crc_wait;
            
            end if;
        elsif txd_word_ready = '0' and txd_word_valid = '1' then
           -- txd_word_ready <= '0';
            tx_fsm <= data;
            --tx_nibble_counter <= (others => '0');
        end if;
        when crc_wait =>
            txd_word_ready <= '0';
            if (mac_tx_ready_i = '1' and mac_tx_valid_o = '1') then
              mac_tx_valid_o <= '0';
            end if;  
            tx_nibble_counter <= (others => '0');
            tx_fsm <= crc;
          
        when crc =>
         
            if (mac_tx_valid_o = '0' or (mac_tx_ready_i = '1' and mac_tx_valid_o = '1')) then
              mac_tx_valid_o <= '1';    
            case (to_integer (tx_nibble_counter (2 downto 0))) is
              when 0 => mac_txd_o <= crc_tx_reversed (3 downto 0);
              when 1 => mac_txd_o <= crc_tx_reversed (7 downto 4);
              when 2 => mac_txd_o <= crc_tx_reversed (11 downto 8);
              when 3 => mac_txd_o <= crc_tx_reversed (15 downto 12);
              when 4 => mac_txd_o <= crc_tx_reversed (19 downto 16);
              when 5 => mac_txd_o <= crc_tx_reversed (23 downto 20);
              when 6 => mac_txd_o <= crc_tx_reversed (27 downto 24);
              when 7 => mac_txd_o <= crc_tx_reversed (31 downto 28);
                        mac_tx_eof_o <= '1';
                        tx_fsm <= gap;
                        
            when others =>
                        mac_txd_o <= (others => 'X');
            end case;
          end if;
          if (mac_tx_valid_o = '0' or (mac_tx_ready_i = '1' and mac_tx_valid_o = '1')) then
             tx_nibble_counter <= tx_nibble_counter + 1;
          end if;
         when gap =>
           if (mac_tx_ready_i = '1') then
             mac_tx_valid_o <= '0';
             mac_tx_eof_o <= '0';
             tx_nibble_counter <= (others => '0');
          end if;
          if (mac_tx_valid_o = '0') then
            tx_nibble_counter <= tx_nibble_counter + 1;
            if (tx_nibble_counter = x"F") then
              tx_fsm <= idle;
            end if;
          end if;
         when others =>
            tx_fsm <= idle;
            mac_tx_valid_o <= '0';
            txd_word_ready <= '0';
            end case;
     end if;
     end if;
end process;            

 crc_tx_reversed <= not reverse_vector (crc_tx);
rst_crc_tx <= '1' when (rst_i = '1' or tx_fsm = idle) else '0';   
en_crc_tx <= mac_tx_valid_o and mac_tx_ready_i when (tx_fsm = data or tx_fsm = fetch_data or tx_fsm = crc_wait) else '0';
inst_crc_tx : entity work.ueth_crc
  port map(
    clk_i => clk_i,
    rst_i => rst_crc_tx,

    en_i => en_crc_tx,
    data_i => mac_txd_o,
    crc_o => crc_tx
    );

rst_crc_rx <= '1' when (rst_i = '1' or rx_fsm /= data) else '0';   

          
inst_crc_rx : entity work.ueth_crc
  port map(
    clk_i => clk_i,
    rst_i => rst_crc_rx,

    en_i => mac_rx_valid_i and not mac_rx_eof_i,
    data_i => mac_rxd_i,
    crc_o => crc_rx
    );


inst_dma : entity work.ueth_dma

  
  port map (
    clk_i => clk_i,
    rst_i => rst_i,

    rst_mac_o      => rst_mac,
    rx_en_o        => rx_en,
    tx_en_o        => tx_en,
    rx_done_i      => rx_done,
    tx_done_i      => tx_done,
    rx_overflow_i  => rx_overflow,
    tx_underflow_i => tx_underflow,

    tx_pending_o => tx_pending,
   
    rx_ready_o => rx_ready,
    tx_fifo_data_o  => tx_fifo_data_dma,
    tx_fifo_valid_o => tx_fifo_valid_dma,
    tx_fifo_ready_i => tx_fifo_ready_dma,
   

    rx_fifo_data_i  => rx_fifo_data_dma,
    rx_fifo_valid_i => rx_fifo_valid_dma,
    rx_fifo_ready_o => rx_fifo_ready_dma,
    rx_fifo_eof_i => rx_fifo_eof_dma,

    irq_o => irq_o,

    mii_rd_o => mii_rd,
    mii_done_i => mii_done,
    mii_reg_i => mii_reg,
    mii_addr_o => mii_addr,

    
    slave_i => slave_i,
    slave_o => slave_o,

    master_o => master_o,
    master_i => master_i

    );


     fifo_rx : entity work.axis_fifo_cdc
    
    generic map (
        dwidth         => 36,
        uwidth         => 0,
        log2depth      => 4,
        tcq            => 1 )
    port map (
        s_clk          => clk_i,
        s_reset        => rst_i or rst_mac,

        m_clk          => clk_i,
        m_reset        => rst_i or rst_mac,
        
        s_axis_data    => rxd_word,
        s_axis_valid   => rxd_word_valid,
        s_axis_user    => (others => '0'),
        s_axis_last    => rxd_word_eof,
        s_axis_ready   => rxd_word_ready,
  
        m_axis_data    => rx_fifo_data_dma,
        m_axis_valid   => rx_fifo_valid_dma,
        m_axis_user    => open,
        m_axis_last    => rx_fifo_eof_dma,
        m_axis_ready   => rx_fifo_ready_dma);
          

     fifo_tx : entity work.axis_fifo_cdc
    
    generic map (
        dwidth         => 36,
        uwidth         => 0,
        log2depth      => 4,
        tcq            => 1 )
    port map (
        s_clk          => clk_i,
        s_reset        => rst_i or rst_mac,

        m_clk          => clk_i,
        m_reset        => rst_i or rst_mac,
        
        s_axis_data    => tx_fifo_data_dma,
        s_axis_valid   => tx_fifo_valid_dma,
        s_axis_user    => (others => '0'),
        s_axis_last    => '0',
        s_axis_ready   => tx_fifo_ready_dma,
   
        m_axis_data    => txd_word,
        m_axis_valid   => txd_word_valid,
        m_axis_user    => open,
        m_axis_last    => open,
        m_axis_ready   => txd_word_ready);

          
inst_eth_mii : entity work.ueth_miim

  port map (
        clk_i => clk_i,
        rst_i => rst_i,

        read_i => mii_rd,
        addr_i => mii_addr,
        done_o      => mii_done,
        reg_o => mii_reg,    

        phy_mdc_o => phy_mdc_o,
        phy_md_i  => phy_md_i,
        phy_md_o  => phy_md_o,
        phy_oe_o  => phy_oe_o
    );


end architecture;

