
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--library synplify;
--use synplify.attributes.all;
--use synplify.all;

library work;
use work.wishbone_pkg.all;
--
--library smartfusion2;
--use smartfusion2.all;


entity ueth_phy is

  
  port (
  
   --TX
    phy_tx_clk_i : in  std_logic;
    rst_tx_i    : in std_logic;
    phy_txd_o    : out std_logic_vector (3 downto 0);
    phy_txen_o   : out std_logic;
    phy_txerr_o  : out std_logic;

    --RX
    phy_rx_clk_i : in std_logic;
    rst_rx_i     : in std_logic;
    phy_rxd_i    : in std_logic_vector (3 downto 0);
    phy_rxdv_i   : in std_logic;
    phy_rxerr_i  : in std_logic;
    col_i        : in std_logic;
    crs_i        : in std_logic;

    tx_pending_i : in std_logic;
    rx_pending_i : in std_logic;

    tx_done_o : out std_logic;
    rx_done_o : out std_logic;
    rx_err_o : out std_logic;
    

    tx_fifo_data_i      : in  std_logic_vector (35 downto 0);
    tx_fifo_valid_i     : in  std_logic;
    tx_fifo_ready_o     : out std_logic;
    tx_fifo_underflow_o : out std_logic;
    tx_fifo_empty_i     : in std_logic;
    tx_byte_sel_i : in std_logic_vector (3 downto 0);

    rx_fifo_data_o     : out std_logic_vector (35 downto 0);
    rx_fifo_valid_o    : out std_logic;
    rx_fifo_ready_i    : in  std_logic;
    rx_fifo_overflow_o : out std_logic;
    rx_fifo_empty_i : in std_logic;
    rx_byte_sel_o : out std_logic_vector (3 downto 0)


    );
end entity ueth_phy;


architecture rtl of ueth_phy is

type t_tx_fsm is (idle, preamble, sfd, data, crc, gap, wait_ack); 
type t_rx_fsm is (idle, wait_preamble, wait_sfd, data, wait_ack);

signal tx_fsm : t_tx_fsm := idle;
signal rx_fsm : t_rx_fsm := idle;

signal preamble_counter, gap_counter : unsigned (3 downto 0);
signal tx_nibble_counter, rx_nibble_counter : unsigned (2 downto 0);

signal rx_data, crc_tx, crc_rx : std_logic_vector (31 downto 0);
signal rst_tx, rst_rx, rx_done, rx_done_d : std_logic;

signal tx_data : std_logic_vector (35 downto 0);

signal enable_crc_tx : std_logic;
signal phy_txd_n,phy_rxd_n : std_logic_vector (3 downto 0);

function reverse_vector (a: in std_logic_vector)
return std_logic_vector is
  variable result: std_logic_vector(a'RANGE);
  alias aa: std_logic_vector(a'REVERSE_RANGE) is a;
begin
  for i in aa'RANGE loop
    result(i) := aa(i);
  end loop;
  return result;
end; -- function reverse_any_vector

function f_word2nibble (data : std_logic_vector (31 downto 0); nibble : unsigned (2 downto 0)) return std_logic_vector is
  variable data_rev : std_logic_vector (31 downto 0);
begin
  data_rev := not reverse_vector (data);
  
case (to_integer(nibble)) is
  when 7 => return (data_rev(31 downto 28));
  when 6 => return (data_rev(27 downto 24));
  when 5 => return (data_rev(23 downto 20));
  when 4 => return  (data_rev(19 downto 16));
  when 3 => return  (data_rev(15 downto 12));
  when 2 => return  (data_rev(11 downto 8));
  when 1 => return  (data_rev(7 downto 4));
  when 0 => return (data_rev(3 downto 0)); 
when others => return "XXXX"; 
end case;
end f_word2nibble;



function f_byte_sel (byte_counter : unsigned (1 downto 0)) return std_logic_vector is
begin
  case (to_integer (byte_counter)) is
    when 0 => return "1111";
    when 1 => return "0001";
    when 2 => return "0011";
    when 3 => return "0111";
    when others => return "XXXX";
  end case;
end f_byte_sel;


  function f_byte_count (byte_sel : std_logic_vector (3 downto 0)) return integer is
  begin
  case (byte_sel) is
    when "0001" => return 1;
    when "0011" => return 2;
    when "0111" => return 3;
    when "1111" => return 4;
    when others => return -1;
  end case;
  end f_byte_count;
 

  attribute syn_encoding : string; 
attribute syn_encoding of tx_fsm, rx_fsm : signal is "safe,onehot";

attribute syn_radhardlevel : string;
attribute syn_radhardlevel of rtl : architecture is "tmr";

begin



  
  process (phy_tx_clk_i, rst_tx)
  begin
  if (rst_tx_i = '1') then
    
    
        tx_fsm          <= idle;
        tx_done_o       <= '0';
        tx_fifo_ready_o <= '0';
        tx_fifo_underflow_o <= '0';
 phy_txen_o            <= '0';
      elsif rising_edge(phy_tx_clk_i) then
        
        case (tx_fsm) is
          when idle =>
            tx_done_o <= '0';
            phy_txd_o <= (others => '0');
            phy_txerr_o           <= '0';
            phy_txen_o            <= '0';
            enable_crc_tx         <= '0';
            preamble_counter <= (others => '0');
            tx_nibble_counter <= (others => '0');
            gap_counter       <= (others => '0');
            
            if (tx_fifo_valid_i = '1' and tx_pending_i = '1') then
              tx_fsm <= preamble;
              tx_fifo_underflow_o <= '0';
            end if;
          when preamble =>
            phy_txd_o  <= x"5";
            phy_txen_o <= '1';
            preamble_counter    <= preamble_counter + 1;
            if (preamble_counter = 15) then
              tx_fsm <= sfd;
            end if;
          when sfd =>
            phy_txd_o       <= x"D";
            tx_fsm          <= data;
            tx_data         <= tx_fifo_data_i; 
            tx_fifo_ready_o <= '1';
            enable_crc_tx <= '1';
          when data =>
            
            tx_fifo_ready_o       <= '0';
            phy_txd_o  <= tx_data (3 downto 0);
            tx_data (31 downto 0) <= "0000" & tx_data (31 downto 4);
            tx_nibble_counter     <= tx_nibble_counter + 1;
            if (tx_data (35 downto 32) /= "1111" and tx_nibble_counter = 2*f_byte_count (tx_fifo_data_i (35 downto 32))-1) then
              tx_fsm <= crc;
              enable_crc_tx <= '0';
              tx_nibble_counter <= (others => '0');
            end if;
            if (tx_fifo_valid_i = '0' and tx_nibble_counter = 7 ) then
              if ( tx_pending_i = '1') then
                tx_fifo_underflow_o <= '1';
                 phy_txerr_o         <= '1';
                 phy_txd_o             <= (others => '1');
            else
              
              tx_fsm          <= crc;
              enable_crc_tx <= '0';
              tx_nibble_counter <= (others => '0');
            end if;
          end if;
            if (tx_nibble_counter = 7 and tx_fifo_valid_i = '1') then
              tx_data <= tx_fifo_data_i;
              tx_fifo_ready_o <= '1';
          end if;
              
--            case (to_integer(tx_nibble_counter)) is
--                when 7 => phy_txd_o <= tx_fifo_data_i (31 downto 28);
--                when 6 => phy_txd_o <= tx_fifo_data_i (27 downto 24);
--                when 5 => phy_txd_o <= tx_fifo_data_i (23 downto 20);
--                when 4 => phy_txd_o <= tx_fifo_data_i (19 downto 16);
--                when 3 => phy_txd_o <= tx_fifo_data_i (15 downto 12);
--                when 2 => phy_txd_o <= tx_fifo_data_i (11 downto 8);
--                when 1 => phy_txd_o <= tx_fifo_data_i (7 downto 4);
--                when 0 => phy_txd_o <= tx_fifo_data_i (3 downto 0);
--              when others => phy_txd_o <= (others => 'X');
--               end case;
           -- if (tx_nibble_counter = 6 and tx_fifo_empty_i = '0') then
--               tx_fifo_ready_o <= '1';
--            end if;
--            if(tx_fifo_data_i (35 downto 32) /= "1111" and tx_nibble_counter = 2*f_byte_count (tx_fifo_data_i (35 downto 32))-1) then
--              tx_fifo_ready_o <= '1';
--              tx_fsm          <= crc;
--              tx_nibble_counter <= (others => '0');
--            end if;
--            if (tx_fifo_empty_i = '1'  and tx_fifo_ready_o = '0' and tx_nibble_counter = 7) then
--              
--              if (tx_pending_i = '1') then
--              --to do error
--              tx_fifo_underflow_o <= '1';
--              phy_txerr_o         <= '1';
--              phy_txd_o             <= (others => '1');
--              
--              else
--              tx_fifo_ready_o <= '1';
--              tx_fsm          <= crc;
--              tx_nibble_counter <= (others => '0');
--              end if;
--             end if;
            
          when crc =>
            enable_crc_tx <= '0';
            tx_fifo_ready_o <= '0';
                   
          
            phy_txerr_o <= '0';
            phy_txd_o   <= f_word2nibble(crc_tx, tx_nibble_counter);
            tx_nibble_counter <= tx_nibble_counter + 1;
            if (tx_nibble_counter = 7) then
              tx_fsm <= gap;
            end if;
          when gap =>
            phy_txen_o  <= '0';
            phy_txd_o <= (others => '0');
            gap_counter <= gap_counter + 1;
            if (gap_counter = 11) then
              tx_fsm    <= idle;
              tx_done_o <= '1';
            end if;
        when wait_ack =>


          when others =>
              tx_fsm <= idle;
        end case;
      end if;
    --end if;
  end process;


phy_txd_n <=  tx_data (3 downto 0);

  process (phy_tx_clk_i)
  begin
    if rising_edge (phy_tx_clk_i) then
      if (tx_fsm = preamble) then
        crc_tx <= (others => '1');
      elsif (enable_crc_tx = '1') then
        crc_tx(0)  <= crc_tx(28) xor phy_txd_n(3);
        crc_tx(1)  <= crc_tx(28) xor crc_tx(29) xor phy_txd_n(2) xor phy_txd_n(3);
        crc_tx(2)  <= crc_tx(28) xor crc_tx(29) xor crc_tx(30) xor phy_txd_n(1) xor phy_txd_n(2) xor phy_txd_n(3);
        crc_tx(3)  <= crc_tx(29) xor crc_tx(30) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(1) xor phy_txd_n(2);
        crc_tx(4)  <= crc_tx(0) xor crc_tx(28) xor crc_tx(30) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(1) xor phy_txd_n(3);
        crc_tx(5)  <= crc_tx(1) xor crc_tx(28) xor crc_tx(29) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(2) xor phy_txd_n(3);
        crc_tx(6)  <= crc_tx(2) xor crc_tx(29) xor crc_tx(30) xor phy_txd_n(1) xor phy_txd_n(2);
        crc_tx(7)  <= crc_tx(3) xor crc_tx(28) xor crc_tx(30) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(1) xor phy_txd_n(3);
        crc_tx(8)  <= crc_tx(4) xor crc_tx(28) xor crc_tx(29) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(2) xor phy_txd_n(3);
        crc_tx(9)  <= crc_tx(5) xor crc_tx(29) xor crc_tx(30) xor phy_txd_n(1) xor phy_txd_n(2);
        crc_tx(10) <= crc_tx(6) xor crc_tx(28) xor crc_tx(30) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(1) xor phy_txd_n(3);
        crc_tx(11) <= crc_tx(7) xor crc_tx(28) xor crc_tx(29) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(2) xor phy_txd_n(3);
        crc_tx(12) <= crc_tx(8) xor crc_tx(28) xor crc_tx(29) xor crc_tx(30) xor phy_txd_n(1) xor phy_txd_n(2) xor phy_txd_n(3);
        crc_tx(13) <= crc_tx(9) xor crc_tx(29) xor crc_tx(30) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(1) xor phy_txd_n(2);
        crc_tx(14) <= crc_tx(10) xor crc_tx(30) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(1);
        crc_tx(15) <= crc_tx(11) xor crc_tx(31) xor phy_txd_n(0);
        crc_tx(16) <= crc_tx(12) xor crc_tx(28) xor phy_txd_n(3);
        crc_tx(17) <= crc_tx(13) xor crc_tx(29) xor phy_txd_n(2);
        crc_tx(18) <= crc_tx(14) xor crc_tx(30) xor phy_txd_n(1);
        crc_tx(19) <= crc_tx(15) xor crc_tx(31) xor phy_txd_n(0);
        crc_tx(20) <= crc_tx(16);
        crc_tx(21) <= crc_tx(17);
        crc_tx(22) <= crc_tx(18) xor crc_tx(28) xor phy_txd_n(3);
        crc_tx(23) <= crc_tx(19) xor crc_tx(28) xor crc_tx(29) xor phy_txd_n(2) xor phy_txd_n(3);
        crc_tx(24) <= crc_tx(20) xor crc_tx(29) xor crc_tx(30) xor phy_txd_n(1) xor phy_txd_n(2);
        crc_tx(25) <= crc_tx(21) xor crc_tx(30) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(1);
        crc_tx(26) <= crc_tx(22) xor crc_tx(28) xor crc_tx(31) xor phy_txd_n(0) xor phy_txd_n(3);
        crc_tx(27) <= crc_tx(23) xor crc_tx(29) xor phy_txd_n(2);
        crc_tx(28) <= crc_tx(24) xor crc_tx(30) xor phy_txd_n(1);
        crc_tx(29) <= crc_tx(25) xor crc_tx(31) xor phy_txd_n(0);
        crc_tx(30) <= crc_tx(26);
        crc_tx(31) <= crc_tx(27);

      end if;
    end if;
  end process;

  process (phy_rx_clk_i, rst_rx)
  begin
if (rst_rx_i = '1') then
    
      
        rx_fsm          <= idle;
  
        rx_done_o     <= '0';
        rx_fifo_valid_o <= '0';
        rx_err_o        <= '0';
        rx_fifo_overflow_o <= '0';
      elsif rising_edge(phy_rx_clk_i) then

        case (rx_fsm) is
          when idle =>
            rx_done_o <= '0';
            rx_err_o <= '0';
            rx_data <= (others => '0');
            rx_fifo_valid_o <= '0';
            rx_nibble_counter <= (others => '0');
            rx_byte_sel_o <= (others => '0');
            if (rx_pending_i = '1') then
              rx_done_o <= '0';
              rx_err_o <= '0';
              rx_fifo_overflow_o <= '0';
              rx_fsm <= wait_preamble; 
            end if;
        when wait_preamble =>
            if (phy_rxdv_i = '1' and phy_rxd_i = x"5") then
              rx_fsm <= wait_sfd;

            end if;
          when wait_sfd =>
            if (phy_rxd_i = x"D") then
              rx_fsm <= data;
            end if;

          when data =>
            rx_fifo_valid_o <= '0';
            if (phy_rxdv_i = '1' ) then
             -- rx_data           <= rx_data (27 downto 0) & phy_rxd_i;
              rx_nibble_counter <= rx_nibble_counter + 1;
              case (to_integer(rx_nibble_counter)) is
                when 7 => rx_fifo_data_o (31 downto 28) <= phy_rxd_i;
                when 6 => rx_fifo_data_o (27 downto 24) <= phy_rxd_i;
                when 5 => rx_fifo_data_o (23 downto 20) <= phy_rxd_i;
                when 4 => rx_fifo_data_o (19 downto 16) <= phy_rxd_i;
                when 3 => rx_fifo_data_o (15 downto 12) <= phy_rxd_i;
                when 2 => rx_fifo_data_o (11 downto 8) <= phy_rxd_i;
                when 1 => rx_fifo_data_o (7 downto 4) <= phy_rxd_i;
                when 0 => rx_fifo_data_o (3 downto 0)   <= phy_rxd_i;
              when others => rx_fifo_data_o <= (others => 'X');
               end case;
            
              end if;
           if ( (rx_nibble_counter = 7)) then
              rx_fifo_valid_o <= '1';
              rx_fifo_data_o (35 downto 32) <= "1111";
         end if;
         
              if (phy_rxdv_i = '0' and rx_nibble_counter /= 0) then
                rx_fifo_data_o (35 downto 32) <= f_byte_sel(rx_nibble_counter (2 downto 1));
                rx_fifo_valid_o <= '1';
              end if;
            if (phy_rxdv_i = '0' ) then

              rx_fsm    <= wait_ack;
              rx_nibble_counter <= (others => '0');
            end if;

            if (rx_fifo_ready_i = '0') then
              rx_fifo_overflow_o <= '1';
            end if;
            
           if (phy_rxdv_i = '1'  and phy_rxerr_i = '1') then
             rx_err_o <= '1';
        end if;
     
         when wait_ack =>
          rx_fifo_valid_o <= '0';
          if (rx_fifo_empty_i = '1') then
          rx_done_o <= '1';
          end if;
          if (rx_pending_i = '0') then
            rx_fsm <= idle;
        end if; 
            
          when others =>
            rx_fsm <= idle;
        end case;

        if (phy_rxdv_i = '1' and phy_rxerr_i = '1') then
          rx_err_o <= '1';
        end if;
      end if;
    --end if;
  end process;


phy_rxd_n <=  phy_rxd_i;
 process (phy_rx_clk_i,crc_rx, phy_rxd_n)
  begin
    if rising_edge (phy_rx_clk_i) then
      if (rx_fsm = wait_sfd) then
        crc_rx <= (others => '1');
      elsif (rx_fsm = data and phy_rxdv_i = '1') then
              crc_rx(0)  <= crc_rx(28) xor phy_rxd_n(3);
        crc_rx(1)  <= crc_rx(28) xor crc_rx(29) xor phy_rxd_n(2) xor phy_rxd_n(3);
        crc_rx(2)  <= crc_rx(28) xor crc_rx(29) xor crc_rx(30) xor phy_rxd_n(1) xor phy_rxd_n(2) xor phy_rxd_n(3);
        crc_rx(3)  <= crc_rx(29) xor crc_rx(30) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(1) xor phy_rxd_n(2);
        crc_rx(4)  <= crc_rx(0) xor crc_rx(28) xor crc_rx(30) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(1) xor phy_rxd_n(3);
        crc_rx(5)  <= crc_rx(1) xor crc_rx(28) xor crc_rx(29) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(2) xor phy_rxd_n(3);
        crc_rx(6)  <= crc_rx(2) xor crc_rx(29) xor crc_rx(30) xor phy_rxd_n(1) xor phy_rxd_n(2);
        crc_rx(7)  <= crc_rx(3) xor crc_rx(28) xor crc_rx(30) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(1) xor phy_rxd_n(3);
        crc_rx(8)  <= crc_rx(4) xor crc_rx(28) xor crc_rx(29) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(2) xor phy_rxd_n(3);
        crc_rx(9)  <= crc_rx(5) xor crc_rx(29) xor crc_rx(30) xor phy_rxd_n(1) xor phy_rxd_n(2);
        crc_rx(10) <= crc_rx(6) xor crc_rx(28) xor crc_rx(30) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(1) xor phy_rxd_n(3);
        crc_rx(11) <= crc_rx(7) xor crc_rx(28) xor crc_rx(29) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(2) xor phy_rxd_n(3);
        crc_rx(12) <= crc_rx(8) xor crc_rx(28) xor crc_rx(29) xor crc_rx(30) xor phy_rxd_n(1) xor phy_rxd_n(2) xor phy_rxd_n(3);
        crc_rx(13) <= crc_rx(9) xor crc_rx(29) xor crc_rx(30) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(1) xor phy_rxd_n(2);
        crc_rx(14) <= crc_rx(10) xor crc_rx(30) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(1);
        crc_rx(15) <= crc_rx(11) xor crc_rx(31) xor phy_rxd_n(0);
        crc_rx(16) <= crc_rx(12) xor crc_rx(28) xor phy_rxd_n(3);
        crc_rx(17) <= crc_rx(13) xor crc_rx(29) xor phy_rxd_n(2);
        crc_rx(18) <= crc_rx(14) xor crc_rx(30) xor phy_rxd_n(1);
        crc_rx(19) <= crc_rx(15) xor crc_rx(31) xor phy_rxd_n(0);
        crc_rx(20) <= crc_rx(16);
        crc_rx(21) <= crc_rx(17);
        crc_rx(22) <= crc_rx(18) xor crc_rx(28) xor phy_rxd_n(3);
        crc_rx(23) <= crc_rx(19) xor crc_rx(28) xor crc_rx(29) xor phy_rxd_n(2) xor phy_rxd_n(3);
        crc_rx(24) <= crc_rx(20) xor crc_rx(29) xor crc_rx(30) xor phy_rxd_n(1) xor phy_rxd_n(2);
        crc_rx(25) <= crc_rx(21) xor crc_rx(30) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(1);
        crc_rx(26) <= crc_rx(22) xor crc_rx(28) xor crc_rx(31) xor phy_rxd_n(0) xor phy_rxd_n(3);
        crc_rx(27) <= crc_rx(23) xor crc_rx(29) xor phy_rxd_n(2);
        crc_rx(28) <= crc_rx(24) xor crc_rx(30) xor phy_rxd_n(1);
        crc_rx(29) <= crc_rx(25) xor crc_rx(31) xor phy_rxd_n(0);
        crc_rx(30) <= crc_rx(26);
        crc_rx(31) <= crc_rx(27);



      end if;
    end if;
  end process;



end architecture;

