library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.wishbone_pkg.all;


entity wb_envm is

 port (
    clk_i 			: in std_logic;
    rst_i   		: in std_logic;
	rst_cpu_o 	: out std_logic;  
	
    apb_psel_o      : out std_logic;
	apb_pwr_o		: out std_logic; 
	apb_pen_o 		: out std_logic;
	apb_paddr_o 	: out std_logic_vector(31 downto 0);
	apb_pwdat_o 	: out std_logic_vector(31 downto 0);
	apb_prdat_i		: in std_logic_vector(31 downto 0);
	apb_pready_i	: in std_logic; --
	
    -- to the Wishbone xbar
    master_i : in  t_wishbone_master_in;
    master_o : out t_wishbone_master_out;
	
    slave_i : in  t_wishbone_slave_in;
    slave_o : out t_wishbone_slave_out

    );
end entity wb_envm;


architecture rtl of wb_envm is

constant bootloader_addr : std_logic_vector (31 downto 0) := x"60012000"; -- (bits from 11 to 0 must be 0)
constant bootloader_size : integer := 256; -- words (max 256)

type fsm_apb_states is (idle, setup, accessing); 
type fsm_envm_states is (idle, check_status, poll_bit, data); 
type fsm_bootstrap_states is (cold_start, poll_bit, fetch_word, wait_apb, wait_wb, checksum, done);
type fsm_wb_master_states is (idle, wait_ack);

signal apb_fsm : fsm_apb_states := idle;
signal envm_fsm : fsm_envm_states := idle;
signal bootstrap_fsm : fsm_bootstrap_states := cold_start;
signal wb_fsm : fsm_wb_master_states := idle;

signal apb_addr, apb_addr_boot, apb_addr_wb, envm_addr, envm_data, apb_data, wb_data, wb_data_reg : std_logic_vector (31 downto 0);

signal envm_read_req_p, apb_read_req, apb_read_req_boot, apb_read_req_wb, apb_read_ack : std_logic;
signal wb_write_req, wb_write_req_boot, wb_write_req_wb, wb_write_ack : std_logic;
signal wb_addr, wb_addr_boot, wb_addr_wb : std_logic_vector (16 downto 0);
signal counter : unsigned (9 downto 0);
begin

apb_addr <= apb_addr_boot when (rst_cpu_o = '1') else apb_addr_wb;
apb_read_req <= apb_read_req_boot when (rst_cpu_o = '1') else apb_read_req_wb;

process (clk_i)
begin
if rising_edge (clk_i) then
	if rst_i = '1' then
		rst_cpu_o <= '1';
		bootstrap_fsm <= cold_start;
		wb_write_req_boot <= '0';
	else
		case (bootstrap_fsm) is
			when cold_start =>
				rst_cpu_o <= '1';
				bootstrap_fsm <= poll_bit;
				counter <= (others => '0');
				wb_write_req_boot <= '0';
				apb_addr_boot <=  x"60080120";
				apb_read_req_boot <= '1';
			when poll_bit =>
				apb_read_req_boot <= '0';
				if (apb_read_ack = '1') then
					if (apb_data (0) = '1') then	-- busy bit
						bootstrap_fsm <= fetch_word;
					else
						apb_read_req_boot <= '1';
					end if;
				end if;
			when fetch_word =>
				apb_read_req_boot <= '1';
				apb_addr_boot (31 downto 12) <= bootloader_addr (31 downto 12);
				apb_addr_boot (11 downto 2) <= std_logic_vector(counter);
				apb_addr_boot (1 downto 0) <= (others => '0');
				bootstrap_fsm <= wait_apb;
			when wait_apb =>
				apb_read_req_boot <= '0';
				if (apb_read_ack = '1') then
					bootstrap_fsm <= wait_wb;
					wb_write_req_boot <= '1';				
					wb_addr_boot <= (others => '0');
					wb_addr_boot (11 downto 2) <= std_logic_vector(counter);		
				end if;
			when wait_wb =>
				wb_write_req_boot <= '0';
				if (wb_write_ack = '1') then
					if (counter >= to_unsigned(bootloader_size-4, counter'length)) then
						bootstrap_fsm <= checksum;
					else
						bootstrap_fsm <= fetch_word;
						counter <= counter + 1;
					end if;
				end if;
			when checksum =>
				-- to be done
				bootstrap_fsm <= done;
			when done =>
				rst_cpu_o <= '0';
			when others =>
				bootstrap_fsm <= done;
		end case;
	end if;
end if;
end process;

wb_addr <= wb_addr_boot when (rst_cpu_o = '1') else wb_addr_wb;
wb_write_req <= wb_write_req_boot when (rst_cpu_o = '1') else wb_write_req_wb;
wb_data <= apb_data when (rst_cpu_o = '1') else wb_data_reg;


process (clk_i)
begin
if rising_edge (clk_i) then
	if rst_i = '1' then
		wb_fsm <= idle;
		master_o.cyc <= '0';
		master_o.stb <= '0';
		master_o.adr <= (others => '0');
	else
		case (wb_fsm) is
			when idle =>
				master_o.cyc <= '0';
				master_o.stb <= '0';
				wb_write_ack <= '0';
				if (wb_write_req = '1' and wb_write_ack = '0') then
					wb_fsm <= wait_ack;
					master_o.cyc <= '1';
					master_o.stb <= '1';
				    master_o.we                <= '1';
					master_o.sel               <= "1111";				
					master_o.adr (wb_addr'left downto 0) <= wb_addr; 
					master_o.dat               <= wb_data;
				end if;
			when wait_ack =>
				if (master_i.ack = '1') then
					master_o.cyc <= '0';
					master_o.stb <= '0';
					master_o.we <= '0';
					wb_write_ack <= '1';
					wb_fsm <= idle;
				end if;
			when others =>
				wb_fsm <= idle;
		end case;
					
					
	end if;
end if;
end process;

process (clk_i)
begin
if rising_edge(clk_i) then
	if rst_i = '1' then
		apb_fsm <= idle;
		apb_psel_o <= '0';
		apb_pen_o <= '0';
		apb_pwr_o <= '0';
        apb_read_ack <= '0';
	else 
		case (apb_fsm) is 
			when idle =>
                apb_read_ack <= '0';
				if (apb_read_req = '1' and apb_read_ack = '0') then
					apb_fsm <= setup;
					apb_psel_o <= '1';
					apb_paddr_o <= apb_addr;
				end if;
			when setup =>
				apb_pen_o <= '1';
				apb_fsm <= accessing;
			when accessing =>
				if (apb_pready_i = '1') then
					apb_data <= apb_prdat_i;
					apb_fsm <= idle;
					apb_pen_o <= '0';
					apb_psel_o <= '0';
					apb_read_ack <= '1';
				end if;
			when others =>
				apb_fsm <= idle;
		end case;
	end if;
end if;
end process;

process (clk_i) 
begin
if rising_edge (clk_i) then
	if rst_i = '1' then
		envm_fsm <= idle;
		apb_read_req_wb <= '0';
	else
	case (envm_fsm) is
		when idle =>
			if (envm_read_req_p = '1') then
				envm_fsm <= check_status;
			end if;
		when check_status =>
			if (apb_fsm = idle) then
				apb_addr_wb <= x"60080120";
				apb_read_req_wb <= '1';
				envm_fsm <= poll_bit;
			end if;
		when poll_bit =>
			apb_read_req_wb <= '0';
			if (apb_read_ack = '1') then
				if (apb_data (0) = '1') then	-- busy bit
					envm_fsm <= data;
					apb_addr_wb <= envm_addr;
					apb_read_req_wb <= '1';
				else
					apb_read_req_wb <= '1';		-- retry 
				end if;
			end if;
		when data =>
			apb_read_req_wb <= '0';
			if (apb_read_ack = '1') then
				envm_data <= apb_data;
				envm_fsm <= idle;
			end if;
		when others =>
			envm_fsm <= idle;
	end case;
end if;
end if;
end process;
			
process (clk_i)
begin
if rising_edge (clk_i) then
	envm_read_req_p <= '0';
	wb_write_req_wb <= '0';
    slave_o.ack <= '0';
	if (slave_i.stb = '1' and slave_i.cyc = '1' and slave_o.ack = '0') then
        slave_o.ack <= '1';
		if (slave_i.adr(4 downto 2) = "000") then
			slave_o.dat (31 downto 1) <= (others => '0');
			slave_o.dat (0) <= '1' when envm_fsm /= idle else '0';
		elsif (slave_i.adr (4 downto 2) = "001") then
			slave_o.dat <= slave_i.dat;
			if (slave_i.we = '1') then
				envm_addr <= slave_i.dat;
				envm_read_req_p <= '1';
			end if;
		elsif (slave_i.adr (4 downto 2) = "010") then
			slave_o.dat <= envm_data;
		elsif (slave_i.adr (4 downto 2) = "100") then
			slave_o.dat <= slave_i.dat;
			if (slave_i.we = '1') then
				wb_addr_wb <= slave_i.dat (wb_addr_wb'left downto 0);			
			end if;
		elsif (slave_i.adr (4 downto 2) = "101") then
			slave_o.dat <= slave_i.dat;		
			if (slave_i.we = '1') then
				wb_data_reg <= slave_i.dat;
				wb_write_req_wb <= '1';
			end if;
		end if;
	end if;
end if;
end process;	

		

		
			


end architecture;