
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
--library synplify;
--use synplify.attributes.all;
--use synplify.all;

library work;
use work.wishbone_pkg.all;
--library smartfusion2;
--use smartfusion2.all;


entity xwb_ram_adapter is
  generic (
    g_addr_width : natural := 15;
    g_data_width :natural := 39;
    g_is_iram : boolean := false;
    g_is_dram : boolean := false

    );
  port (
    clk_i : in std_logic;
    rst_i : in std_logic;
    boot_done_i : in std_logic;

    -- Wishbone
    slave_i : in  t_wishbone_slave_in;
    slave_o : out t_wishbone_slave_out;

    -- ram interface
    d_ram_i      : in  std_logic_vector (g_data_width-1 downto 0);
    q_ram_o      : out std_logic_vector (g_data_width-1 downto 0);
    a_ram_o      : out std_logic_vector (g_addr_width-1 downto 0);
    we_ram_o     : out std_logic;
    re_ram_o     : out std_logic;
    valid_ram_i  : in  std_logic;
    lock_req_o   : out std_logic;
    lock_grant_i : in  std_logic;
    
          single_error_p_o : out std_logic;
      double_error_p_o : out std_logic

    );
end entity xwb_ram_adapter;


architecture rtl of xwb_ram_adapter is

  component secded_ecc is
    generic (
      g_addr_width : natural := 15
      );
    port (
      clk_i : in std_logic;
      rst_i : in std_logic;


      -- to the processor/bus
      d_i   : in std_logic_vector(31 downto 0);
      a_i   : in std_logic_vector(g_addr_width-1 downto 0);
      we_i  : in std_logic;
      bwe_i : in std_logic_vector (3 downto 0);
      re_i  : in std_logic;

      q_o      : out std_logic_vector(31 downto 0);
      done_r_o : out std_logic;
      done_w_o : out std_logic;

      --to the BRAM
      d_ram_i      : in  std_logic_vector (38 downto 0);
      q_ram_o      : out std_logic_vector (38 downto 0);
      a_ram_o      : out std_logic_vector (g_addr_width-1 downto 0);
      we_ram_o     : out std_logic;
      re_ram_o     : out std_logic;
      valid_ram_i  : in  std_logic;
      lock_req_o   : out std_logic;
      lock_grant_i : in  std_logic;

      single_error_p_o : out std_logic;
      double_error_p_o : out std_logic;
      read_fault_p_o   : out std_logic;
      write_fault_p_o  : out std_logic
      );


  end component;
  
    component secded_ecc_daec is
    generic (
      g_addr_width : natural := 15
      );
    port (
      clk_i : in std_logic;
      rst_i : in std_logic;


      -- to the processor/bus
      d_i   : in std_logic_vector(31 downto 0);
      a_i   : in std_logic_vector(g_addr_width-1 downto 0);
      we_i  : in std_logic;
      bwe_i : in std_logic_vector (3 downto 0);
      re_i  : in std_logic;

      q_o      : out std_logic_vector(31 downto 0);
      done_r_o : out std_logic;
      done_w_o : out std_logic;

      --to the BRAM
      d_ram_i      : in  std_logic_vector (41 downto 0);
      q_ram_o      : out std_logic_vector (41 downto 0);
      a_ram_o      : out std_logic_vector (g_addr_width-1 downto 0);
      we_ram_o     : out std_logic;
      re_ram_o     : out std_logic;
      valid_ram_i  : in  std_logic;
      lock_req_o   : out std_logic;
      lock_grant_i : in  std_logic;

      single_error_p_o : out std_logic;
      double_error_p_o : out std_logic;
      read_fault_p_o   : out std_logic;
      write_fault_p_o  : out std_logic
      );


  end component;


  signal addr_scrubber : unsigned (15 downto 0);
  signal timeout_wb : unsigned (3 downto 0);
  signal pwr, wr       : std_logic;

  type fsm_states is (idle, process_wb, scrubber_read);
  signal fsm : fsm_states := idle;

 signal we, re : std_logic;
 signal addr : std_logic_vector (g_addr_width-1 downto 0);
 signal data_in, data_out : std_logic_vector (31 downto 0);
signal bwe : std_logic_vector (3 downto 0);

  attribute syn_radhardlevel        : string;
  attribute syn_radhardlevel of rtl : architecture is "tmr";


  signal ack, done_r, done_w : std_logic;

  signal data_read : std_logic_vector (31 downto 0);

attribute syn_encoding : string; 
attribute syn_encoding of fsm : signal is "safe,onehot";


begin

 
  slave_o.ack   <= ack;
  slave_o.stall <= (not ack) and slave_i.cyc and slave_i.stb;
  slave_o.err <= '0';
  slave_o.int <= '0';
   slave_o.rty <= '0'; 
  process (clk_i)
  begin

    if rising_edge (clk_i) then
    if rst_i = '1' then
      fsm           <= idle;
      addr_scrubber <= (x"0000");
      we         <= '0';
      re         <= '0';
      ack           <= '0';
      timeout_wb <= (others => '0');
      addr <= (others => '0');
    else
      
      if (timeout_wb /= (timeout_wb'range => '1')) then
        timeout_wb <= timeout_wb + 1;
      end if;

      case (fsm) is
        when idle =>
          we <= '0';
          re <= '0';
          ack   <= '0';
          if (slave_i.stb = '1' and slave_i.cyc = '1' and ack = '0') then
            fsm <= process_wb;
            we <= slave_i.we;
            
            if boot_done_i = '0' then
                we  <= slave_i.we;
            elsif (g_is_iram and slave_i.we = '1') then
                we <= '0';
                fsm <= idle;
                report "Attempt to write iram in runtime" severity error;
            elsif (g_is_dram and slave_i.we = '1') then
              if (g_addr_width = 15 and slave_i.adr (g_addr_width-1 downto g_addr_width-3) /= "111") then
                we <= '0';
                fsm <= idle;
                report "Attempt to write lower part of dram in runtime" severity error;
              end if;
            
               if (g_addr_width = 16 and slave_i.adr (g_addr_width-1 downto g_addr_width-4) /= "1111") then
                we <= '0';
                fsm <= idle;
                report "Attempt to write lower part of dram in runtime" severity error;
              end if;
            end if;           
            bwe <= slave_i.sel;
            re  <= not slave_i.we;
            data_in <= slave_i.dat;
            addr <= slave_i.adr (g_addr_width-1 downto 0);
          elsif (boot_done_i = '1' and timeout_wb = (timeout_wb'range => '1')) then
            fsm <= scrubber_read;
            re <= '1';
            addr <= std_logic_vector(addr_scrubber(g_addr_width-1 downto 0));
          end if;
        when process_wb =>

          slave_o.dat <= data_out;

          if ((re = '1' and done_r = '1') or (we = '1' and done_w = '1')) then
              re <= '0';
              we <= '0';
              fsm <= idle;
              ack <= '1';
              timeout_wb <= (others => '0');
          end if;
       
   
        when scrubber_read =>
        
          if (done_r = '1') then
            re <= '0';
            fsm           <= idle;
            addr_scrubber <= addr_scrubber+4;
          end if;
        when others =>
          fsm <= idle;
      end case;
    end if;
end if;
  end process;

ecc_simple : if g_data_width = 39 generate
  ecc_circuitry : secded_ecc
    generic map (
      g_addr_width => g_addr_width
      )
  port map (
    clk_i => clk_i,
    rst_i => rst_i,


    d_i  => data_in,
    a_i  => addr,
    we_i => we and not done_w,
    bwe_i => bwe,
    re_i => re and not done_r,

    q_o      => data_out,
    done_r_o => done_r,
    done_w_o => done_w,

    --to the BRAM
    d_ram_i     => d_ram_i,
    q_ram_o     => q_ram_o,
    a_ram_o     => a_ram_o,
    we_ram_o    => we_ram_o,
    re_ram_o    => re_ram_o,
    valid_ram_i => valid_ram_i,

      lock_req_o  => lock_req_o,
      lock_grant_i => lock_grant_i,

    single_error_p_o => single_error_p_o,
    double_error_p_o => double_error_p_o,
    read_fault_p_o   => open,
    write_fault_p_o  => open
    );
end generate ecc_simple;

ecc_daec : if g_data_width = 42 generate
 ecc_circuitry : secded_ecc_daec
    generic map (
      g_addr_width => g_addr_width
      )
  port map (
    clk_i => clk_i,
    rst_i => rst_i,


    d_i  => data_in,
    a_i  => addr,
    we_i => we and not done_w,
    bwe_i => bwe,
    re_i => re and not done_r,

    q_o      => data_out,
    done_r_o => done_r,
    done_w_o => done_w,

    --to the BRAM
    d_ram_i     => d_ram_i,
    q_ram_o     => q_ram_o,
    a_ram_o     => a_ram_o,
    we_ram_o    => we_ram_o,
    re_ram_o    => re_ram_o,
    valid_ram_i => valid_ram_i,

      lock_req_o  => lock_req_o,
      lock_grant_i => lock_grant_i,

    single_error_p_o => single_error_p_o,
    double_error_p_o => double_error_p_o,
    read_fault_p_o   => open,
    write_fault_p_o  => open
    );

end generate ecc_daec;


end architecture;